import logging
import gc

from .ioutils import (
    get_resolutons,
    check_weight,
    store_weights
)
from .core import (
    balance_ic,
    balance_kr
)

def main(args):
    for resolution in get_resolutons(args.mcool):
        cooleruri = args.mcool + '::resolutions/' + resolution

        for weight_name, per_chrom, balancetype in zip(
            ['KR', 'perChromKR'],
            [False, True],
            ['genomewide', 'per chromosome']
        ):
            if not check_weight(cooleruri, weight_name) or args.overwrite:
                logging.info(
                    'applying {} KR to {}::resolution/{}'.format(
                        balancetype, 
                        args.mcool, 
                        resolution
                    )
                )
                krweights = balance_kr(
                    cooleruri, 
                    per_chrom
                )
                store_weights(
                    cooleruri, 
                    krweights, 
                    weight_name,
                    overwrite = args.overwrite
                )
                del krweights

            else:
                logging.info(
                    '{} KR weights for {}::resolution/{} already exist. Skipping!'.format(
                        balancetype, 
                        args.mcool, 
                        resolution
                    )
                )

        for weight_name, per_chrom, balancetype in zip(
            ['ICE', 'perChromIC'], 
            [False, True], 
            ['genomewide', 'per chromosome']
        ):
            if not check_weight(cooleruri, weight_name) or args.overwrite:
                logging.info(
                    'applying {} IC to {}::resolution/{}'.format(
                        balancetype, 
                        args.mcool, 
                        resolution
                    )
                )
                icweights, stats = balance_ic(
                    cooleruri, 
                    args.processors, 
                    per_chrom
                )
                store_weights(
                    cooleruri, 
                    icweights, 
                    weight_name,
                    stats,
                    overwrite = args.overwrite
                )
                del icweights

            else:
                logging.info(
                    '{} IC weights for {}::resolution/{} already exist. Skipping!'.format(
                        balancetype, 
                        args.mcool, 
                        resolution
                    )
                )

        gc.collect()



import glob
import re
import os

import pandas as pd

from cooler import Cooler


cs_regex = re.compile('_(?P<cs>[0-9]+)_')


def clustersize_from_filename(filename):
    m = cs_regex.search(filename)
    return int(m.group('cs'))


def read_coolers(directory):
    coolers = {}
    for coolfile in glob.glob(directory + '/*'):
        clustersize = clustersize_from_filename(
            os.path.basename(coolfile)
        )
        coolers[clustersize] = Cooler(coolfile)
    
    return coolers



import pandas as pd
import numpy as np
import itertools as it

from functools import partial


# all this could surely also be done by iterating over pixels
# however this needs additional overhead as pixels selector does not respect bin ids
# so one has to ensure that the pixel frames are aligned which is way easier with matrix


def merge_pixels(coolers, rowlo, rowhi, collo, colhi, round_floats):
    pixels = pd.DataFrame()
    tmp_count_cols = ['count_x', 'count_y']
    for size, cool in coolers.items():
        selector = cool.matrix(balance = False, as_pixels = True)
        tmp = selector[rowlo:rowhi, collo:colhi]
        if tmp.empty:
            continue

        tmp['count'] = tmp['count'].astype(float)
        tmp.loc[:, 'count'] = tmp['count'] * (2/size)

        if pixels.empty:
            pixels = tmp
            continue
        
        pixels = pixels.merge(
            tmp,
            on = ['bin1_id', 'bin2_id'],
            how = 'outer'
        )
        pixels['count'] = pixels[tmp_count_cols].sum(axis = 1)
        pixels.drop(
            columns = tmp_count_cols,
            inplace = True
        )
    
    # in case all pixel frames are empty
    if pixels.empty:
        return pixels

    if round_floats:
        pixels['float_count'] = pixels['count']
        pixels.loc[:, 'count'] = pixels.float_count.round(0)

    return pixels


def get_nbins(coolers):
    if isinstance(coolers, dict):
        key = np.random.choice(list(coolers.keys()))
        cool = coolers[key]

    cool = coolers
    return cool.info['nbins']


def chunked_pixels_iterator(pixels_selector, coolers, nchunks = 100, round_floats = True):
    chunkextents = np.linspace(
        0, 
        get_nbins(coolers), 
        nchunks, 
        dtype = int
    )
    for rowlo, rowhi in it.pairwise(chunkextents):
        for collo, colhi in it.pairwise(chunkextents):
            pixels = pixels_selector(
                coolers,
                rowlo, rowhi,
                collo, colhi,
                round_floats
            )
            if pixels.empty:
                continue

            yield pixels


def pixels_from_matrix_block(cooler, rowlo, rowhi, collo, colhi, round_floats):
    pixels = cooler.matrix(balance = False, as_pixels = True)[rowlo:rowhi, collo:colhi]

    if round_floats:
        pixels['float_count'] = pixels['count']
        pixels.loc[:, 'count'] = pixels.float_count.round(0)

    return pixels


chunked_pixels_merge = partial(
    chunked_pixels_iterator,
    merge_pixels
)

chunked_pixels_block = partial(
    chunked_pixels_iterator,
    pixels_from_matrix_block
)



import logging

import numpy as np

from .ioutils import read_coolers
from .core import chunked_pixels_merge
from cooler import create_cooler


def main(args):
    logging.info('reading coolers')
    coolers = read_coolers(args.input)
    logging.info('merging coolers')
    if not args.floatcounts:
        columns = ['count', 'float_count']
        dtypes = None
        logging.info('storing float counts in extra column')
    
    else:
        columns = None
        dtypes = {'count': float}
        logging.info('storing count as float')

    key = np.random.choice(list(coolers.keys()))
    bins = coolers[key].bins()[:]

    create_cooler(
        args.outfile,
        bins,
        chunked_pixels_merge(
            coolers, 
            args.nchunks, 
            not args.floatcounts
        ),
        columns = columns,
        dtypes = dtypes,
        ordered = False
    )


from pybedtools import BedTool
from .ioutils import read_bed_by_chrom

import pandas as pd

import os


def sanitize_dtypes(df, dtypes):
    for col, dtype in dtypes.items():
        df[col] = df[col].astype(dtype)


def annotate_bins(cool, clusterbedfile):
    cool_chroms = set(cool.chromnames)
    annotated_bins = []
    for chrom_bed in read_bed_by_chrom(clusterbedfile):
        chrom = chrom_bed.iloc[0, 0]
        if not chrom in cool_chroms:
            continue
        
        lo, hi = cool.extent(chrom)
        chrom_bins = cool.bins()[lo:hi].loc[:, ['chrom', 'start', 'end']]
        chrom_bins['name'] = 'bin'
        a = BedTool.from_dataframe(chrom_bed)
        b = BedTool.from_dataframe(chrom_bins)
        intersect = b.intersect(a, wao = True, sorted = True).to_dataframe()
        grouped = intersect[['chrom', 'start', 'end', 'thickEnd']].groupby(['chrom', 'start', 'end'])
        annotated_bins.append(
            grouped.agg({'thickEnd': ','.join}).reset_index()
        )
    
    tmp_annotation = pd.concat(annotated_bins)
    colname = os.path.basename(clusterbedfile)
    tmp_annotation.rename(
        columns = {'thickEnd': colname},
        inplace = True
    )
    bins = cool.bins()[:].loc[:, ['chrom', 'start', 'end']]
    annotated_bins = bins.merge(
        tmp_annotation, 
        on = ['chrom', 'start', 'end'], 
        how = 'left'
    )
    annotated_bins.fillna({colname: 'None'}, inplace = True)
    sanitize_dtypes(
        annotated_bins,
        {
            'chrom': 'category',
            colname: 'bytes'
        }
    )
    return annotated_bins



import logging

import pandas as pd

from cooler import Cooler, fileops
from .core import annotate_bins
from .ioutils import copy_and_annotate_cooler


def annotate_cool(coolpath, bedpaths, outfile, mcoolfile = False):
    cooler = Cooler(coolpath)

    annotated_bins = pd.DataFrame()
    for bedpath in bedpaths:
        logging.info(f'annotating bins of {coolpath} with clusters from {bedpath}')
        tmp = annotate_bins(cooler, bedpath)
        if annotated_bins.empty:
            annotated_bins = tmp
            continue

        annotated_bins = annotated_bins.merge(
            tmp,
            on = ['chrom', 'start', 'end'],
            how = 'left'
        )

    annotated_bins.drop(
        columns = ['chrom', 'start', 'end'],
        inplace = True
    )
    logging.info(f'writing annotated data to {outfile}')
    copy_and_annotate_cooler(
        coolpath,
        outfile,
        annotated_bins,
        mcoolfile = mcoolfile
    )


def annotate_mcool(mcoolpath, bedpaths, outfile):
    for coolpath in fileops.list_coolers(mcoolpath):
        uri = mcoolpath + '::' + coolpath
        outuri = outfile + '::' + coolpath
        annotate_cool(
            uri,
            bedpaths,
            outuri,
            mcoolfile = True
        )


def main(args):
    for coolpath in args.input:
        if fileops.is_multires_file(coolpath):
            logging.info('annotating multires cooler')
            outfile = coolpath.replace('mcool', 'annotated.mcool')
            annotate_mcool(coolpath, args.bed, outfile)

        else:
            logging.info('annotating single cooler')
            outfile = coolpath.replace('.cool', '.annotated.cool')
            annotate_cool(coolpath, args.bed, outfile)
